/*****************************************************************************
 * $RCSfile: gouraud.cc,v $
 * Copyright 2001,2002 Matthias Seekamp, Andreas Daniel, Daniel Stephan,
 *                     Jens Meyer, Robert Adolph
 * 
 * This file is part of BUGSAgent2002
 *
 * BUGSAgent2002 is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * BUGSAgent2002 is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BUGSAgent2002; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *-CVS-Info:-----------------------------------------------------------------
 * Currently locked by: $Locker:  $
 *
 * Last changed:
 * ------------------------------
 * $Author: fastjack $
 * $Date: 2002/03/25 16:14:07 $
 * $Revision: 1.9 $
 *****************************************************************************/


#include "gouraud.hh"
#include "CommunicationManager.hh"
#include "Log.hh"
#include <cmath>
#include <string.h>


class World;

GouraudFiller::GouraudFiller(int * f, int dx, int dy): field(f), dimx(dx), dimy(dy), fieldsize(dx*dy),
						       fillMode(FM_ADD), initializedSinCos(false) {
  ledge  = new int[dy];
  redge = new int[dy];
  lcolor = new int[dy];
  rcolor = new int[dy];  
}

GouraudFiller::~GouraudFiller() {
  delete [] ledge;
  delete [] redge;
  delete [] lcolor;
  delete [] rcolor;
}

void GouraudFiller::clearArrays(int from, int len) {
  if (len == -1) len = ( dimy - from );
  
  memset( ledge+from, 0, sizeof(int)*len );
  memset( redge+from, 0, sizeof(int)*len );
  memset( lcolor+from, 0, sizeof(int)*len );
  memset( rcolor+from, 0, sizeof(int)*len );
}
  
float GouraudFiller::int_pow(float x, int p)
{
  if (p < 0) 
    return (1.0 / int_pow(x,-p));
  else {
    float ans = 1.0;
    for (int i=0; i<p; i++)
      ans *= x;
    return ans;
  }
}

float GouraudFiller::round(float x, int p)
{
  x *= int_pow(10.0, -p);
  if (fmod(x, 1.0) >= .5)
    return ceil(x) / int_pow(10.0, -p);
  else
    return floor(x) / int_pow(10.0, -p);
} 

void GouraudFiller::buildCosineTable() {
  //
  float r;
  float d2rFactor = M_PI/180;
  for (int d=0; d<360; d++) {
    r = d * d2rFactor;
    cost[d] = cos(r);
  }
}
void GouraudFiller::buildSineTable() {
  //
  float r;
  float d2rFactor = M_PI/180;
  for (int d=0; d<360; d++) {
    r = d * d2rFactor;
    sint[d] = sin(r);
  }
}

/** is fast but not accurate */
void GouraudFiller::rotatePointAroundPoint(int &x,int &y,
					   int px,int py,
					   int degree) {
  int newx,newy;
  x = x-px;
  y = y-py;
  newx = int(cost[degree] * x - sint[degree] * y);
  newy = int(cost[degree] * y + sint[degree] * x);
  x = newx+px;
  y = newy+py;
  
}

int GouraudFiller::getValue(int x, int y) {
  if ((y >= dimy)||(x >= dimx)) return -1;
  return *(int *)(y*dimx + x);
}
void GouraudFiller::setValue(int v, int x, int y) {
  //    Pruefung muss vorher erfolgen (scanEdge)
  
  int dp = y*dimx + x;
  int * pInt = field + dp;

  if ((dp >= fieldsize)||(dp < 0)) {
    Log::debug << "Gouraud ERROR, try to write out of bounds (" <<x<<","<<y<<") = " <<v<< endl;
    return;
  }
  
  switch (fillMode) {
  case FM_REPLACE:  *pInt = v; break;
  case FM_ADD:      *pInt = (*pInt) + v; break;
  case FM_MULTIPLY: *pInt = (*pInt) * v; break;
  case FM_OR:       *pInt = (*pInt) | v; break;
  case FM_XOR:      *pInt = (*pInt) ^ v; break;
  case FM_AND:      *pInt = (*pInt) & v; break;
  default: break;
  }
  
}

void GouraudFiller::fillHLine(int x1, int x2, int y, int v1, int v2) {
  
  if (x2<x1) { // make sure that x1 < x2
    int tmp = x1;
    x1 = x2;
    x2 = tmp;
    tmp = v1; // have to switch the color also ...
    v1 = v2;
    v2 = tmp;
  }
  
  if (x1==x2) {
    if ((x1>=0)&&(x1<dimx))
      setValue(v1,x1,y);
    return;
  }
  
  int mv  = (v2 - v1) << 8;
  mv /= (x2 - x1);
  
  int v = v1 << 8;
  
  if (x1 < 0) x1 = 0;
  if (x2 >= dimx) x2=dimx-1;
  
  for (int i=x1;i<x2;i++) {
    setValue(v >> 8,i,y);
    v += mv;
  }

}


void GouraudFiller::scanEdge(int *edge, int *color,
			     int x1, int x2, int y1, int y2, int c1, int c2) {
  
  // Lines ausserhalb des Feldes ignorieren
  
  if (y1 == y2) return;

  if ((y1 < 0)&&(y2 < 0)) return;
  if ((y1 >= dimy)&&(y2 >= dimy)) return;

  
  if (y1 > y2) {
    int tmp = y1;
    y1=y2;
    y2=tmp;

    tmp=x1;
    x1=x2;
    x2=tmp;

    tmp=c1;
    c1=c2;
    c2=tmp;
  }
  
  if (y1 > y2) Log::info << "GouraudFiller: y1 is bigger than y2!" << endl;
  
  int mx  = (x2 - x1) << 16;
  mx /= (y2 - y1);
  
  int mc  = (c2 - c1) << 16;
  mc /= (y2 - y1);
  
  int x = x1 << 16;             // scale the starting x value
  int c = c1 << 16;             // scale the starting c value
  
  // y1 kann nicht ueber dimy liegen, weil dann auch y2 ueber dimy waere und dann
  // schon vorher rausgesprungen waeren
  // y2 kann nicht unter 0 liegen, weil dann y1 auch unter 0 waere und wir daher
  // schon vorher rausgesprungen waeren
  if (y1<0) {
    // oh, teil liegt ausserhalb des feldes. *panik*
    x1 += mx * (-y1);
    c1 += mc * (-y1);
    y1 = 0;
  }
  if (y2>dimy) {
    // dito. *panik*
    y2=dimy;
  }
  
  int v,cv;
  
  for (int count = y1; count < y2; count++)
    {
      v = x >> 16;
      cv = c >> 16;
      
      edge[count] = v;
      color[count] = cv;
      
      x += mx;
      c += mc;
    }
}

void GouraudFiller::gouraudFillTriangle(int x1,int x2,int x3,
					int y1,int y2,int y3,
					int c1,int c2,int c3) {

  int maxy=(y1<y2)?((y2<y3)?y3:y2):((y1<y3)?y3:y1);
  int miny=(y1<y2)?((y1<y3)?y1:y3):((y2<y3)?y2:y3);
  
  // Falls y Werte komplett ausserhalb des Feldes liegen, nix fuellen.
  if ((maxy <0) && (miny<0)) return;
  if ((maxy >=dimy)&&(miny >=dimy)) return;
  
  // minx und maxx merken, um spaeter genau diese zeilen zu fuellen.
  if (miny<0) miny = 0;
  if (maxy>=dimy) maxy = dimy-1;
  
  int maxx=(x1<x2)?((x2<x3)?x3:x2):((x1<x3)?x3:x1);
  int minx=(x1<x2)?((x1<x3)?x1:x3):((x2<x3)?x2:x3);
  
  // Falls x Werte komplett ausserhalb des Feldes liegen, nix fuellen.
  if ((maxx <0) && (minx<0)) return;
  if ((maxx >=dimx)&&(minx >=dimx)) return;
  
  clearArrays(miny,maxy-miny);

  //
  // es werden alle Kanten des Dreiecks gescannt und
  // deren Eckwerte in zwei arrays geschrieben. Eines
  // fr die linke Seite und eines fr die rechte Seite -
  // ebenso gibt es zwei arrays fr die "Farb"-Werte.
  // die ganzen berprfungen sind dafr da, dass die
  // lngste Kante in das ledge array, und die anderen
  // beiden in das redge array eingetragen werden knnen.
  // Ebenso soll abgefangen werden, falls eine Kante
  // waagerecht verluft. Dann gibts nmlich eine Division
  // durch 0 in der scanEdge Funktion.
  // Wenn die Kante waagerecht verluft, tragen wir sie
  // einfach direkt in die arrays ein.
  
  // irgendwie gibt es hier einen Gedankenfehler.
  // wenn man die lngste Dreiecksseite in das ledge
  // eintrgt heisst das ja noch nicht, dass damit die linke
  // Begrenzung vollstndig gescannt wurde.
  // Dennoch werden ja alle krzeren Kanten ins redge
  // eingetragen.
  //
  
  // so sortieren, dass (x1,y1)-(x2,y2) die groesste 
  // y-Differenz aufweist
  sortValues(x1,y1,c1,
	     x2,y2,c2,
	     x3,y3,c3);
  
  scanEdge(ledge,lcolor,x1,x2,y1,y2,c1,c2);
  scanEdge(redge,rcolor,x2,x3,y2,y3,c2,c3);
  scanEdge(redge,rcolor,x3,x1,y3,y1,c3,c1);
 
  
  // jetzt malen wir die Werte ins Feld.
  for (int count = miny; count < maxy; count++)
    fillHLine(ledge[count],redge[count], count, lcolor[count], rcolor[count]);
}

int GouraudFiller::int_abs(int a) {
  if (a<0) return -a;
  return a;
}

void GouraudFiller::sortValues(int &x1, int &y1, int &c1,
			       int &x2, int &y2, int &c2,
			       int &x3, int &y3, int &c3) {
  int dy1 = int_abs(y1-y2);
  int dy2 = int_abs(y2-y3);
  int dy3 = int_abs(y3-y1);
  
  if ((dy1>=dy2)&&(dy1>=dy3)) return; // schon richtig sortiert gewesen

  int tx,ty,tc; // temp variablen
  
  if ((dy2>=dy3)&&(dy2>=dy3)) {
    tx=x1;
    ty=y1;
    tc=c1;
    x1=x2;
    y1=y2;
    c1=c2;
    x2=x3;
    y2=y3;
    c2=c3;
    x3=tx;
    y3=ty;
    c3=tc;
  } else
  if ((dy3>=dy1)&&(dy3>=dy1)) {
    tx=x1;
    ty=y1;
    tc=c1;
    x1=x3;
    y1=y3;
    c1=c3;
    x3=x2;
    y3=y2;
    c3=c2;
    x2=tx;
    y2=ty;
    c2=tc;
  }  
}


bool GouraudFiller::isInBounds(int x, int y) {
  return !((y < 0)||(y >= dimy)||(x<0)||(x>=dimx));
}  

void GouraudFiller::shadeTriangle( int x1,int y1, int color1,
				   int x2,int y2, int color2,
				   int x3,int y3, int color3 ) {
  gouraudFillTriangle(x1,x2,x3,y1,y2,y3,color1,color2,color3);
}
void GouraudFiller::shadePyramid( int midx,int midy, int halfSideLength, int height ) {
  
  int x1 = midx-halfSideLength;
  int x2 = midx+halfSideLength;
  int y1 = midy-halfSideLength;
  int y2 = midy+halfSideLength;
  
  gouraudFillTriangle(x1,midx,midx,
		      y1,midy, y1,
		      0,height,0);
  gouraudFillTriangle(midx,midx,x2,
		      y1, midy,y1,
		      0,height,0);
  gouraudFillTriangle(x2,midx,x2,
		      y1,midy,midy,
		      0,height, 0);
  gouraudFillTriangle(x2,midx,x2,
		      midy,midy,y2,
		      0,height, 0);
  gouraudFillTriangle(x2,midx,midx,
		      y2,midy,y2,
		      0,height, 0);
  gouraudFillTriangle(midx,midx,x1,
		      y2,midy,y2,
		      0,height, 0);
  gouraudFillTriangle(x1,midx,x1,
		      y2,midy,midy,
		      0,height, 0);
  gouraudFillTriangle(x1,midx, x1,
		      y1,midy,midy,
		      0,height, 0);
  
}

void GouraudFiller::shadeTrapezoid( int x, int y,
				    int widthNear, int widthFar,
				    int length, float degree,
				    int height1,int height2,
				    int height3,int height4) {
  initSinCosTables();
  
  
  int halfNearWidth = widthNear/2;
  int halfFarWidth = widthFar/2;
  int x1 = x-halfNearWidth;
  int x2 = x+halfNearWidth;
  int x3 = x+halfFarWidth;
  int x4 = x-halfFarWidth;
  int y1,y2,y3,y4;
  y1 = y2 = y;
  y3 = y4 = y+length;

  int d = (int)round(degree);
  
  d = d%360;
  
  // that was the easy part - pretending we would live in a world
  // relative to the trapezoid itself. now we have to rotate the
  // coordinates, translating this local coordinate system to the
  // global coordinate system
  // to do this, we have to rotate all of them around the P
  
  
  rotatePointAroundPoint(x1,y1,x,y,d);
  rotatePointAroundPoint(x2,y2,x,y,d);
  rotatePointAroundPoint(x3,y3,x,y,d);
  rotatePointAroundPoint(x4,y4,x,y,d);
  
  /* aus irgendwelchen Gruenden erscheint die Form zur y Achse durch
     den Punkt x,y gespiegelt
     also zurueckspiegeln... ist wahrscheinlich rumdoktern an symptomen
     wer ne bessere idee hat darf sich gerne melden
  */
  
  y1 = y - (y1-y);
  y2 = y - (y2-y);
  y3 = y - (y3-y);
  y4 = y - (y4-y);
  // x - (x1-x) = x - x1 + x
  // 6 - (10-6) = 6 - 10 + 6
  
  gouraudFillTriangle(x1,x2,x3,
		      y1,y2,y3,
		      height1,height2,height3);
  gouraudFillTriangle(x1,x3,x4,
		      y1,y3,y4,
		      height1,height3,height4);
  
}


/*
 * $Log: gouraud.cc,v $
 * Revision 1.9  2002/03/25 16:14:07  fastjack
 * Merged & added Behaviour classes for offensive behaviour
 *
 * Revision 1.8  2002/03/18 17:13:03  fastjack
 * Absturzursache im GouraudFiller behoben.
 *
 * Revision 1.7  2002/03/15 13:26:01  fastjack
 * Ausgaben entfernt / dem Gouraud Kerl mehr Genauigkeit verschafft und einen boesen BUG entfernt
 *
 * Revision 1.6  2002/03/01 11:18:48  fastjack
 * GPL added
 *
 * Revision 1.5  2002/02/18 17:37:17  seekamp
 * Fehler im cone behoben
 *
 * Revision 1.4  2002/02/15 13:51:10  fastjack
 * GouraudFiller repariert, fehlerhaften Algorithmus gefixt. Paar Kommentare entfernt.
 *
 * Revision 1.3  2002/02/14 19:39:54  fastjack
 * Gouraud Shader verbessert - sind noch Probleme drin leider.
 *
 * Revision 1.2  2002/02/12 12:09:18  fastjack
 * Alternative Cone Methode ZUM TESTEN eingefuegt. Bei Nichtgefallen wieder entfernen :-D
 *
 * Revision 1.1  2002/02/12 00:52:01  fastjack
 * new files
 *
 */

